from contextlib import asynccontextmanager

from fastapi import BackgroundTasks, FastAPI, Request, Response
from fastapi.staticfiles import StaticFiles
from requests import request as make_request

from api.optimization import MultiProcessOptimizer
from api.routers import runs

# see https://github.com/tiangolo/fastapi/issues/3091#issuecomment-821522932 and https://github.com/encode/starlette/issues/1094#issuecomment-730346075 for heavy-load computation

DEBUG = True
backend = None


@asynccontextmanager
async def lifespan(app: FastAPI):
    global backend
    # Load the backend (which runs models in a separate process)
    backend = MultiProcessOptimizer(debug=DEBUG)
    with backend:
        # add endpoints from backend
        # TODO allow to get dynamically
        actions = [("/api/action/evolve", backend.optimizer.evolve)]
        for path, target in actions:
            app.add_api_route(path, target)
        app.openapi_schema = None
        app.openapi()

        yield
        # Unload the backend freeing used ressources by the separate process
        # automatically done when with block is exited
        print("Releasing resources")
    # release remaining allocations
    del backend
    # TODO somehow not all ressources are released upon uvicorn reload, need to investigate further..


app = FastAPI(debug=DEBUG, title="Prompt Optimization Backend", lifespan=lifespan)
app.include_router(runs.router, prefix="/api/runs")


# start optimization
@app.get("/api/run")
def run(num_iterations: int, background_tasks: BackgroundTasks) -> str:
    background_tasks.add_task(backend.run_optimization, num_iterations)
    return f"Running optimization with {num_iterations} iterations."


# get progress
@app.get("/api/progress")
def get_run_progress() -> str:
    result = backend.get_progress()
    return result


# get run state
@app.get("/api/status")
async def get_run_status() -> bool:
    return backend._running


# get current genealogy of prompts
@app.get("/api/get_tree")
async def get_family() -> dict:
    return backend.optimizer.family_tree


if DEBUG:

    @app.get("/")
    @app.get("/static/{_:path}")
    async def frontent_dev(request: Request):
        response = make_request(
            method=request.method,
            url=f"{request.url.replace(port=3000)}",
            headers=request.headers,
            stream=True,
        )
        return Response(
            content=response.raw.read(),
            status_code=response.status_code,
            headers=dict(response.headers.items()),
        )

else:
    app.mount("/", StaticFiles(directory="frontend/build/", html=True), name="frontend")
