from argparse import Namespace
from concurrent.futures import ProcessPoolExecutor

from evolution import GeneticAlgorithm
from models import Llama2, LLMModel
from task import SentimentAnalysis

# def create_model():
#     global optimizer
#     optimizer = Optimizer(debug=DEBUG)

_evolution_model: LLMModel | None = None
_evaluation_model: LLMModel | None = None
# _model_call_type: get_type_hints(LLMModel).get("__call__")


def _setup_models() -> None:
    global _evolution_model, _evaluation_model

    if _evolution_model is not None:
        raise Exception("Evolution model has already been initialized.")

    # currently fix model
    options = Namespace(llama_path="./models/llama-2-13b-chat.Q5_K_M.gguf", chat=True)
    _evolution_model = Llama2(options)

    if _evaluation_model is not None:
        raise Exception("Evaluation model has already been initialized.")

    # currently fix model
    _evaluation_model = _evolution_model


def _release_models() -> None:
    global _evolution_model, _evaluation_model
    del _evolution_model
    del _evaluation_model


def _call_evolution_model(*args, **kwargs):
    return _evolution_model(*args, **kwargs)


def _call_evaluation_model(*args, **kwargs):
    return _evaluation_model(*args, **kwargs)


def f():
    pass


class MultiProcessOptimizer:
    _instance: "MultiProcessOptimizer" = None
    _running: bool = False
    model_exec: ProcessPoolExecutor = None

    def __new__(cls, *args, **kwargs):
        # only allow to create one instance (singleton pattern)
        if cls._instance is None:
            cls._instance = super(MultiProcessOptimizer, cls).__new__(cls)
        return cls._instance

    def __init__(self, *, debug: bool = False) -> None:
        # a flag indicating whether optimizer is currently running
        self.debug = debug

    def __enter__(self):
        # TODO allow to customize optimizer

        # create necessary models
        # initialize worker processes; only 1 worker since prediction is memory-intensive
        # since we only have 1 worker we just save the state in the global namespace which the single worker accesses
        self.model_exec = ProcessPoolExecutor(max_workers=1, initializer=_setup_models)
        # make sure that initializer is called
        self.model_exec.submit(f).result()
        evolution_model = lambda *args, **kwargs: self.model_exec.submit(
            _call_evolution_model, *args, **kwargs
        ).result()
        evaluation_model = lambda *args, **kwargs: self.model_exec.submit(
            _call_evaluation_model, *args, **kwargs
        ).result()

        # currently fix task
        options = Namespace(use_grammar=False, debug=self.debug)
        task = SentimentAnalysis(evaluation_model, options)

        optimizer_class = GeneticAlgorithm
        # optimizer_class = DifferentialEvolution

        self.optimizer = optimizer_class(
            population_size=10,
            task=task,
            evolution_model=evolution_model,
            evaluation_model=evaluation_model,
        )

    def __exit__(self, exc_type, exc_value, exc_tb):
        print(f"Shutting down")
        self._submit(_release_models).result()
        self.model_exec.shutdown(False)
        self.model_exec = None

    def _submit(self, fn, *fn_args, **fn_kwargs):
        if self.model_exec is None:
            raise RuntimeError(
                "Cannot access model executor - you have to use this class as a context manager with the with statement first."
            )
        return self.model_exec.submit(fn, *fn_args, **fn_kwargs)

    def run_optimization(self, num_iterations: int) -> str:
        self._running = True
        self.optimizer.run(num_iterations, debug=self.debug, add_snapshot_dict={})
        self._running = False

    def get_progress(self):
        if hasattr(self.optimizer, "iterations_pbar"):
            result = str(self.optimizer.iterations_pbar)
        else:
            result = "Optimization has not run yet."
        return result
